-- The layer6 of Convolutional Deep Learning on FPGA
-- by Toshio Iwata at digitalfilter.com 2022/07/01

library ieee ;
use ieee.std_logic_1164.all ;
use ieee.std_logic_arith.all ;
use ieee.std_logic_signed.all;

-------------------------------------------------------------------------------
-- Entity
-------------------------------------------------------------------------------
entity conv_layer6 is
port(
  CLK : in std_logic;                            -- clock (100MHz)
  RST_N : in std_logic;                          -- reset (active low)
  CO9 : in std_logic_vector(3 downto 0); 
  CnnStart : IN std_logic;
  W6DataSel : in std_logic_vector(5 downto 0);
  W6DataEnb : in std_logic;
  RAM_X5_RD_ADDR : in std_logic_vector(6 downto 0);
  RAM_X5_RD_SEL : in std_logic;
  RAM_X6_WR_ADDR: out std_logic_vector(5 downto 0);
  RAM_X6_WR_N: out std_logic;
  LAY6IN : in std_logic_vector(4 downto 0); 
  LAY6OUT : out std_logic_vector(19 downto 0)); 
end conv_layer6;

-------------------------------------------------------------------------------
-- Architecture
-------------------------------------------------------------------------------
architecture RTL of conv_layer6 is

component ROM_W6
Port (
          CLK : In std_logic;
          ADDR : In std_logic_vector(6 downto 0);
          DATA : Out std_logic_vector(386 downto 0) );
end component;

component dotpro_w6
Port (
  CLK : in std_logic;                            -- clock (100MHz)
  RST_N : in std_logic;                          -- reset (active low)
  PixRomData : in std_logic_vector(4 downto 0); 
  MatRomData : in std_logic_vector(8 downto 0);
  CO9is0 : in std_logic;
  MATMULOUT : Out std_logic_vector(19 downto 0) );  
end component;

-------------------------------------------------------------------------------
-- Signal
-------------------------------------------------------------------------------
signal W6Data: std_logic_vector(386 downto 0);

signal W6Addr : std_logic_vector(6 downto 0);

signal coeff_00: std_logic_vector(8 downto 0);

signal mac_00: std_logic_vector(19 downto 0);

signal BIAS_00, BIAS_01, BIAS_02, BIAS_03, BIAS_04, BIAS_05, BIAS_06, BIAS_07, BIAS_08, BIAS_09: std_logic_vector(15 downto 0);
signal BIAS_10, BIAS_11, BIAS_12, BIAS_13, BIAS_14, BIAS_15, BIAS_16, BIAS_17, BIAS_18, BIAS_19: std_logic_vector(15 downto 0);
signal BIAS_20, BIAS_21, BIAS_22, BIAS_23, BIAS_24, BIAS_25, BIAS_26, BIAS_27, BIAS_28, BIAS_29: std_logic_vector(15 downto 0);
signal BIAS_30, BIAS_31, BIAS_32, BIAS_33, BIAS_34, BIAS_35, BIAS_36, BIAS_37, BIAS_38, BIAS_39: std_logic_vector(15 downto 0);
signal BIAS_40, BIAS_41, BIAS_42: std_logic_vector(15 downto 0);

signal BIAS_SEL_tmp: std_logic_vector(15 downto 0);
signal BIAS_SEL_00: std_logic_vector(19 downto 0);

signal plusBias_00: std_logic_vector(19 downto 0);

signal Relu_00: std_logic_vector(4 downto 0);

signal CO9is0 : std_logic;
signal CO9is0_sig : std_logic;

signal inch : std_logic_vector(4 downto 0);
signal W6DataSel_dly : std_logic_vector(5 downto 0);
signal W6DataEnb_dly1 : std_logic;
signal W6DataEnb_dly2 : std_logic;

begin

---------------------------------------------------------------------------------
  inch <= LAY6IN;

  W6Addr <= RAM_X5_RD_ADDR;

--  CO9is0_sig <= '1' when RAM_X5_RD_ADDR = "1111111" or CnnStart = '1' else '0';
  CO9is0_sig <= '1' when RAM_X5_RD_ADDR = "0000000" else '0';

  RAM_X6_WR_N <= '0' when W6DataEnb_dly2 = '1' and CO9is0 = '1' else '1';

--------------------------------------------------------------
process( RST_N, CLK )
begin
        if( RST_N = '0' ) then
                RAM_X6_WR_ADDR <= (others => '0');
        elsif( CLK'event and CLK = '1' ) then
                RAM_X6_WR_ADDR <= W6DataSel_dly - 1;
        end if;
end process;

--------------------------------------------------------------
process( RST_N, CLK )
begin
        if( RST_N = '0' ) then
                CO9is0 <= '0';
        elsif( CLK'event and CLK = '1' ) then
                CO9is0 <= CO9is0_sig;
        end if;
end process;

---------------------------------------------------------------------------------
ROM_W6_inst : ROM_W6 port map (
  CLK => CLK,
  ADDR => W6Addr,
  DATA => W6Data );

-------------------------------------------------------------------------------
process( RST_N, CLK )
begin
  if( RST_N = '0' ) then
    W6DataEnb_dly2 <= '0';
    W6DataEnb_dly1 <= '0';
  elsif( CLK'event and CLK = '1' ) then
    W6DataEnb_dly2 <= W6DataEnb_dly1;
    W6DataEnb_dly1 <= W6DataEnb;
  end if;
end process;

process( RST_N, CLK )
begin
  if( RST_N = '0' ) then
    W6DataSel_dly <= (others => '0');
  elsif( CLK'event and CLK = '1' ) then
    W6DataSel_dly <= W6DataSel;
  end if;
end process;

  coeff_00 <=
          W6Data( 386 downto 378 ) when W6DataSel_dly = "000000" else
          W6Data( 377 downto 369 ) when W6DataSel_dly = "000001" else
          W6Data( 368 downto 360 ) when W6DataSel_dly = "000010" else
          W6Data( 359 downto 351 ) when W6DataSel_dly = "000011" else
          W6Data( 350 downto 342 ) when W6DataSel_dly = "000100" else
          W6Data( 341 downto 333 ) when W6DataSel_dly = "000101" else
          W6Data( 332 downto 324 ) when W6DataSel_dly = "000110" else
          W6Data( 323 downto 315 ) when W6DataSel_dly = "000111" else
          W6Data( 314 downto 306 ) when W6DataSel_dly = "001000" else
          W6Data( 305 downto 297 ) when W6DataSel_dly = "001001" else
          W6Data( 296 downto 288 ) when W6DataSel_dly = "001010" else
          W6Data( 287 downto 279 ) when W6DataSel_dly = "001011" else
          W6Data( 278 downto 270 ) when W6DataSel_dly = "001100" else
          W6Data( 269 downto 261 ) when W6DataSel_dly = "001101" else
          W6Data( 260 downto 252 ) when W6DataSel_dly = "001110" else
          W6Data( 251 downto 243 ) when W6DataSel_dly = "001111" else
          W6Data( 242 downto 234 ) when W6DataSel_dly = "010000" else
          W6Data( 233 downto 225 ) when W6DataSel_dly = "010001" else
          W6Data( 224 downto 216 ) when W6DataSel_dly = "010010" else
          W6Data( 215 downto 207 ) when W6DataSel_dly = "010011" else
          W6Data( 206 downto 198 ) when W6DataSel_dly = "010100" else
          W6Data( 197 downto 189 ) when W6DataSel_dly = "010101" else
          W6Data( 188 downto 180 ) when W6DataSel_dly = "010110" else
          W6Data( 179 downto 171 ) when W6DataSel_dly = "010111" else
          W6Data( 170 downto 162 ) when W6DataSel_dly = "011000" else
          W6Data( 161 downto 153 ) when W6DataSel_dly = "011001" else
          W6Data( 152 downto 144 ) when W6DataSel_dly = "011010" else
          W6Data( 143 downto 135 ) when W6DataSel_dly = "011011" else
          W6Data( 134 downto 126 ) when W6DataSel_dly = "011100" else
          W6Data( 125 downto 117 ) when W6DataSel_dly = "011101" else
          W6Data( 116 downto 108 ) when W6DataSel_dly = "011110" else
          W6Data( 107 downto 99 ) when W6DataSel_dly = "011111" else
          W6Data( 98 downto 90 ) when W6DataSel_dly = "100000" else
          W6Data( 89 downto 81 ) when W6DataSel_dly = "100001" else
          W6Data( 80 downto 72 ) when W6DataSel_dly = "100010" else
          W6Data( 71 downto 63 ) when W6DataSel_dly = "100011" else
          W6Data( 62 downto 54 ) when W6DataSel_dly = "100100" else
          W6Data( 53 downto 45 ) when W6DataSel_dly = "100101" else
          W6Data( 44 downto 36 ) when W6DataSel_dly = "100110" else
          W6Data( 35 downto 27 ) when W6DataSel_dly = "100111" else
          W6Data( 26 downto 18 ) when W6DataSel_dly = "101000" else
          W6Data( 17 downto 9 ) when W6DataSel_dly = "101001" else
          W6Data( 8 downto 0 ) when W6DataSel_dly = "101010" else
          (others => '0');

---------------------------------------------------------------------------------
dotpro_00 : dotpro_w6 port map (
  CLK => CLK,  
  RST_N => RST_N, 
  PixRomData => inch,
  MatRomData => coeff_00,
  CO9is0 => CO9is0,
  MATMULOUT => mac_00 );

---------------------------------------------------------------------------------
  BIAS_00 <= CONV_std_logic_vector(-800 , 16);
  BIAS_01 <= CONV_std_logic_vector( 371 , 16);
  BIAS_02 <= CONV_std_logic_vector( 709 , 16);
  BIAS_03 <= CONV_std_logic_vector(-363 , 16);
  BIAS_04 <= CONV_std_logic_vector(-1092 , 16);
  BIAS_05 <= CONV_std_logic_vector( 1196 , 16);
  BIAS_06 <= CONV_std_logic_vector(-609 , 16);
  BIAS_07 <= CONV_std_logic_vector(-710 , 16);
  BIAS_08 <= CONV_std_logic_vector( 1480 , 16);
  BIAS_09 <= CONV_std_logic_vector(-956 , 16);
  BIAS_10 <= CONV_std_logic_vector( 0 , 16);
  BIAS_11 <= CONV_std_logic_vector( 0 , 16);
  BIAS_12 <= CONV_std_logic_vector( 0 , 16);
  BIAS_13 <= CONV_std_logic_vector( 0 , 16);
  BIAS_14 <= CONV_std_logic_vector( 0 , 16);
  BIAS_15 <= CONV_std_logic_vector( 0 , 16);
  BIAS_16 <= CONV_std_logic_vector( 0 , 16);
  BIAS_17 <= CONV_std_logic_vector( 0 , 16);
  BIAS_18 <= CONV_std_logic_vector( 0 , 16);
  BIAS_19 <= CONV_std_logic_vector( 0 , 16);
  BIAS_20 <= CONV_std_logic_vector( 0 , 16);
  BIAS_21 <= CONV_std_logic_vector( 0 , 16);
  BIAS_22 <= CONV_std_logic_vector( 0 , 16);
  BIAS_23 <= CONV_std_logic_vector( 0 , 16);
  BIAS_24 <= CONV_std_logic_vector( 0 , 16);
  BIAS_25 <= CONV_std_logic_vector( 0 , 16);
  BIAS_26 <= CONV_std_logic_vector( 0 , 16);
  BIAS_27 <= CONV_std_logic_vector( 0 , 16);
  BIAS_28 <= CONV_std_logic_vector( 0 , 16);
  BIAS_29 <= CONV_std_logic_vector( 0 , 16);
  BIAS_30 <= CONV_std_logic_vector( 0 , 16);
  BIAS_31 <= CONV_std_logic_vector( 0 , 16);
  BIAS_32 <= CONV_std_logic_vector( 0 , 16);
  BIAS_33 <= CONV_std_logic_vector( 0 , 16);
  BIAS_34 <= CONV_std_logic_vector( 0 , 16);
  BIAS_35 <= CONV_std_logic_vector( 0 , 16);
  BIAS_36 <= CONV_std_logic_vector( 0 , 16);
  BIAS_37 <= CONV_std_logic_vector( 0 , 16);
  BIAS_38 <= CONV_std_logic_vector( 0 , 16);
  BIAS_39 <= CONV_std_logic_vector( 0 , 16);
  BIAS_40 <= CONV_std_logic_vector( 0 , 16);
  BIAS_41 <= CONV_std_logic_vector( 0 , 16);
  BIAS_42 <= CONV_std_logic_vector( 0 , 16);

  BIAS_SEL_tmp(15 downto 0) <= 
          BIAS_00 when W6DataSel_dly = "000001" else
          BIAS_01 when W6DataSel_dly = "000010" else
          BIAS_02 when W6DataSel_dly = "000011" else
          BIAS_03 when W6DataSel_dly = "000100" else
          BIAS_04 when W6DataSel_dly = "000101" else
          BIAS_05 when W6DataSel_dly = "000110" else
          BIAS_06 when W6DataSel_dly = "000111" else
          BIAS_07 when W6DataSel_dly = "001000" else
          BIAS_08 when W6DataSel_dly = "001001" else
          BIAS_09 when W6DataSel_dly = "001010" else
          BIAS_10 when W6DataSel_dly = "001011" else
          BIAS_11 when W6DataSel_dly = "001100" else
          BIAS_12 when W6DataSel_dly = "001101" else
          BIAS_13 when W6DataSel_dly = "001110" else
          BIAS_14 when W6DataSel_dly = "001111" else
          BIAS_15 when W6DataSel_dly = "010000" else
          BIAS_16 when W6DataSel_dly = "010001" else
          BIAS_17 when W6DataSel_dly = "010010" else
          BIAS_18 when W6DataSel_dly = "010011" else
          BIAS_19 when W6DataSel_dly = "010100" else
          BIAS_20 when W6DataSel_dly = "010101" else
          BIAS_21 when W6DataSel_dly = "010110" else
          BIAS_22 when W6DataSel_dly = "010111" else
          BIAS_23 when W6DataSel_dly = "011000" else
          BIAS_24 when W6DataSel_dly = "011001" else
          BIAS_25 when W6DataSel_dly = "011010" else
          BIAS_26 when W6DataSel_dly = "011011" else
          BIAS_27 when W6DataSel_dly = "011100" else
          BIAS_28 when W6DataSel_dly = "011101" else
          BIAS_29 when W6DataSel_dly = "011110" else
          BIAS_30 when W6DataSel_dly = "011111" else
          BIAS_31 when W6DataSel_dly = "100000" else
          BIAS_32 when W6DataSel_dly = "100001" else
          BIAS_33 when W6DataSel_dly = "100010" else
          BIAS_34 when W6DataSel_dly = "100011" else
          BIAS_35 when W6DataSel_dly = "100100" else
          BIAS_36 when W6DataSel_dly = "100101" else
          BIAS_37 when W6DataSel_dly = "100110" else
          BIAS_38 when W6DataSel_dly = "100111" else
          BIAS_39 when W6DataSel_dly = "101000" else
          BIAS_40 when W6DataSel_dly = "101001" else
          BIAS_41 when W6DataSel_dly = "101010" else
          BIAS_42 when W6DataSel_dly = "101011" else
          (others => '0');

--------------------------------------------------------------
process( RST_N, CLK )
begin
        if( RST_N = '0' ) then
                BIAS_SEL_00 <= (others => '0');
        elsif( CLK'event and CLK = '1' ) then
                BIAS_SEL_00 <= BIAS_SEL_tmp(15) & BIAS_SEL_tmp(15) & BIAS_SEL_tmp(15) & BIAS_SEL_tmp(15) & BIAS_SEL_tmp;
        end if;
end process;

--------------------------------------------------------------
  plusBias_00 <= mac_00 + BIAS_SEL_00;

  LAY6OUT <= plusBias_00;

end RTL;

